{******************************************************************************}
{* This software is provided 'as-is', without any express or                  *}
{* implied warranty. In no event will the author be held liable               *}
{* for any damages arising from the use of this software.                     *}
{*                                                                            *}
{* Permission is granted to anyone to use this software for any               *}
{* purpose, including commercial applications, and to alter it                *}
{* and redistribute it freely, subject to the following                       *}
{* restrictions:                                                              *}
{*                                                                            *}
{*   1. The origin of this software must not be misrepresented,               *}
{*      you must not claim that you wrote the original software.              *}
{*      If you use this software in a product, an acknowledgment              *}
{*      in the product documentation would be appreciated but is              *}
{*      not required.                                                         *}
{*                                                                            *}
{*   2. Altered source versions must be plainly marked as such, and           *}
{*      must not be misrepresented as being the original software.            *}
{*                                                                            *}
{*   3. This notice may not be removed or altered from any source             *}
{*      distribution.                                                         *}
{*                                                                            *}
{*                                                                            *}
{* The initial developer is Theo Lustenberger (temp@theo.ch)                  *}
{*                                                                            *}
{******************************************************************************}

// 29.10.2004

unit uAntiAliasControl;

interface

uses Classes, Types, QControls, QGraphics, Qt, XLib;

type

  TXftFont = record
    Asc: integer;
    Desc: integer;
    Height: integer;
    Max_Advance_Width: integer;
    P1, P2: Pointer;
  end;

  PXftFont = ^TXftFont;

  TXRenderColor = record
    Red: word;
    Green: word;
    Blue: word;
    Alpha: word;
  end;

  TXftColor = record
    Pixel: longword;
    Color: TXRenderColor;
  end;

  TXftDraw = record
    Dummy: Pointer;
  end;

  PXftDraw = ^TXftDraw;

  TXGlyphInfo = packed record
    Width: word;
    Height: word;
    X: smallint;
    Y: smallint;
    XOffset: smallint;
    YOffset: smallint;
  end;

  TFontCacheItem = class
    Font: QGraphics.TFont;
    fPfnt: PXftFont;
  end;

  TFontCache = class
  private
    fFontInfoList: TList;
  protected

  public
    constructor Create;
    destructor Destroy; override;
    function GetFont(ClxFont: QGraphics.TFont): PXFtFont;
    procedure AddFont(ClxFont: QGraphics.TFont; XftFont: PXftFont);
  end;


  TAACanvas = class(TCanvas)
  private
    fDisplay: PDisplay;
    fWindow: XID;
    fColormap: longint;
    fVisual: PVisual;
    fPfnt: PXftFont;
    fDrawx: PXftDraw;
    fFtColor: TXFTColor;
    fOnFontChange: TNotifyEvent;
    fFontCache: TFontCache;
    fDontUseAntiAliasing: Boolean;
    fOldFontChangeHandler: TNotifyEvent;
    fContainer: TObject;
    procedure FontChange(Sender: TObject);
    procedure DrawNeeded;
    procedure DoSetFont;
    function CheckAmpers(var Text: WideString; var Pos: TSize): boolean;
    procedure SetDontUseAntialiasing(Value: Boolean);

  protected
    procedure DoUnderStrike(X, Y: Integer; const Text: WideString);
    procedure DoAmperStrike(X, Y: integer; Pos: TSize);
  public
    constructor Create;
    destructor Destroy; override;
    procedure TextOut(X, Y: Integer; const Text: WideString);
    procedure TextRect(Rect: TRect; X, Y: Integer; const Text: WideString;
      TextFlags: Integer = 0);
    function TextExtent(const Text: WideString): TSize;
    function TextHeight(const Text: WideString): Integer;
    function TextWidth(const Text: WideString): Integer;
    function MaxAdvanceWidth: integer;
    property OnFontChange: TNotifyEvent read fOnFontChange write fOnFontChange;
    property DontUseAntiAliasing: Boolean read fDontUseAntiAliasing write SetDontUseAntiAliasing;
  end;

  THackControl = class(TControl)
  public
    function GetPaintDevice: QPaintDeviceH; override;
    procedure AdjustPainter(Painter: QPainterH); override;
  end;

  TAAControlCanvas = class(TAACanvas)
  private
    FControl: THackControl;
    FClipped: Boolean;
    procedure SetControl(AControl: THackControl);
    procedure SetClipped(const Value: Boolean);
  protected
    procedure BeginPainting; override;
    procedure CreateHandle; override;
  public
    destructor Destroy; override;
    procedure FreeHandle;
    procedure StartPaint;
    procedure StopPaint;
    property Control: THackControl read FControl write SetControl;
    property Clipped: Boolean read FClipped write SetClipped;
  end;

  TAABitmap = class;

  TAABitmapCanvas = class(TAACanvas)
  private
    FBitmap: TAABitmap;
  protected
    procedure BeginPainting; override;
    procedure CreateHandle; override;
  public
    constructor Create(Bitmap: TAABitmap);
  end;

  TAABitmap = class(TBitmap)
  private
    FCanvas: TAABitmapCanvas;
    function GetCanvas: TAABitmapCanvas;
  public
    constructor Create; override;
    destructor Destroy; override;
    class function AssignsTo(Dest: TGraphicClass): Boolean; override;
    procedure Assign(Source: TPersistent); override;
    property Canvas: TAABitmapCanvas read GetCanvas;
  end;

  TCustomAAControl = class(TWidgetControl)
  private
    FCanvas: TAAControlCanvas;
  protected
    procedure CreateWidget; override;
    procedure Painting(Sender: QObjectH; EventRegion: QRegionH); override;
    procedure Paint; virtual;
    procedure MaskChanged; override;
    procedure DrawMask(Canvas: TCanvas); virtual;
    procedure UpdateMask;
    property Canvas: TAAControlCanvas read FCanvas;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

const LibName = '';


function XftDrawCreate(display: PDisplay; win: XID; vis: PVisual; colorm: longint): PXftDraw; cdecl;
function XftDrawCreateBitmap (display: PDisplay; bitmap: Pixmap): PXftDraw; cdecl;
procedure XftDrawChange(xftd: PXftDraw; win: XID); cdecl;
procedure XftDrawDestroy(draw: PXftDraw); cdecl;
function XftFontOpenName(display: PDisplay; scr: integer; par3: PChar): PXftFont; cdecl;
procedure XftFontClose(display: PDisplay; fnt: PXftFont); cdecl;
procedure XftDrawString8(draw: PXftDraw; var col: TXftColor; fnt: PXftFont; x, y: integer; txt: PChar; len: integer); cdecl;
procedure XftDrawString16(draw: PXftDraw; var col: TXftColor; fnt: PXftFont; x, y: integer; txt: PWideChar; len: integer); cdecl;
procedure XftDrawStringUtf8(draw: PXftDraw; var col: TXftColor; fnt: PXftFont; x, y: integer; txt: PChar; len: integer); cdecl;
procedure XftTextExtents16(display: PDisplay; fnt: PXftFont; txt: PWideChar; len: integer; var extents: TXGlyphInfo); cdecl;
procedure XftTextExtents8(display: PDisplay; fnt: PXftFont; txt: PChar; len: integer; var extents: TXGlyphInfo); cdecl;
function XftGlyphExists(display: PDisplay; fnt: PXftFont; ch: integer): longbool; cdecl;
procedure XftDrawRect(draw: PXftDraw; var col: TXftColor; x, y: integer; width, height: Cardinal); cdecl;
procedure XftDrawSetClip(draw: PXftDraw; rg: Region); cdecl;

implementation

uses LibC, RTLConsts, Sysutils, QForms, QConsts;

function XIDFromPainter(Painter: QPainterH): XID;
var device: QPaintDeviceH;
begin
  device := QPainter_device(Painter);
  if device <> nil then
    Result := XID(QPaintDevice_handle(device))
  else
    Result := 0;
end;

{TAACanvas}

constructor TAACanvas.Create;
begin
  inherited Create;
  fDontUseAntiAliasing := false;
  fWindow := 0;
  fDrawX := nil;
  fDisplay := Application.Display;
  fVisual := XDefaultVisual(fDisplay, fDisplay.default_screen);
  fColormap := XDefaultColorMap(fDisplay, fDisplay.default_screen);
  fFtColor.color.alpha := Byte(255) shl 8;
  fFontCache := TFontCache.Create;
  fPFnt := nil;
  DoSetFont;
  fOldFontChangeHandler := Font.OnChange;
  Font.OnChange := FontChange;
  //  DontUseAntiAliasing:=true;
end;


procedure TAACanvas.DrawNeeded;
begin
  if (fWindow <> XIDfromPainter(Handle)) then
  begin
    fWindow := XIDfromPainter(Handle);
    if fDrawX <> nil then XftDrawDestroy(fDrawX);
    fDrawX := XftDrawCreate(fDisplay, fWindow, fVisual, fColorMap);
  end;
end;

destructor TAACanvas.Destroy;
begin
  fFontCache.free;
  inherited;
  if fDrawX <> nil then XftDrawDestroy(fDrawX);
end;

procedure TAACanvas.FontChange(Sender: TObject);
begin
  DoSetFont;
  if Assigned(fOnFontChange) then OnFontChange(self);
end;

procedure TAACanvas.SetDontUseAntialiasing(Value: Boolean);
begin
  if fDontUseAntiAliasing <> Value then
  begin
    if Value then
    begin
      Font.OnChange := fOldFontChangeHandler;
    end else
    begin
      fOldFontChangeHandler := Font.OnChange;
      Font.OnChange := FontChange;
    end;
    fDontUseAntiAliasing := Value;
  end;
end;

function TranslateWeight(Weight:integer):String;
begin
Case Weight of
  fwLight: Result:='light';
  fwNormal: Result:='medium';
  fwDemiBold: Result:='demibold';
  fwBold: Result:='bold';
  fwBlack: Result:='black';
  else
  Result:='medium';
  end;
end;

procedure TAACanvas.DoSetFont;
var fontdescr: string;
  TempFnt: PXftFont;
begin
  if not fDontUseAntiAliasing then
  begin
    fFtColor.color.red := Byte(Font.Color) shl 8;
    fFtColor.color.green := Byte(Font.Color shr 8) shl 8;
    fFtColor.color.blue := Byte(Font.Color shr 16) shl 8;

    TempFnt := fFontCache.GetFont(Font);
    if TempFnt = nil then
    begin
      fontdescr := Font.Name + '-' + inttostr(Font.Size) + ':antialias=true';
      if FsBold in Font.Style then fontdescr := fontdescr + ':bold';
      if FsItalic in Font.Style then fontdescr := fontdescr + ':italic';
      fontdescr := fontdescr + ':weight='+TranslateWeight(Font.Weight);
      if Font.Pitch <> fpDefault then
      begin
       if Font.Pitch = fpFixed then fontdescr := fontdescr + ':spacing=100' else
         if Font.Pitch = fpVariable then fontdescr := fontdescr + ':spacing=0';
      end;
      fpFnt := XftFontOpenName(fDisplay, XDefaultScreen(fDisplay), Pchar(fontdescr));
      fFontCache.AddFont(Font, fpFnt);
    end else
    begin
      fpFnt := TempFnt;
    end;
  end;
end;

function TAACanvas.TextExtent(const Text: WideString): TSize;
var extents: TXGlyphInfo;
begin
  if Text <> '' then
  begin
    if (FDisplay <> nil) and (fpFnt <> nil) and (not fDontUseAntiAliasing) then
    begin
      XftTextExtents16(fDisplay, fpFnt, PWideChar(Text), Length(Text), extents);
      Result.cx := extents.xoffset;
      Result.cy := fpFnt.Height;
    end else Result := inherited TextExtent(Text);
  end else
  begin
    Result.cx := 1;
    Result.cy := 1;
  end;
end;

function TAACanvas.TextHeight(const Text: WideString): Integer;
begin
  if (fpFnt <> nil) and (not fDontUseAntiAliasing) then
    Result := fpFnt.Height else
    Result := Font.Height;
end;

function TAACanvas.TextWidth(const Text: WideString): Integer;
begin
  Result := TextExtent(Text).cx;
end;

function TAACanvas.MaxAdvanceWidth: integer;
begin
  if (fpFnt <> nil) then
  Result:=fpFnt.Max_Advance_Width else Result:=1;
end;

procedure TAACanvas.DoUnderStrike(X, Y: Integer; const Text: WideString);
begin
  if fsUnderLine in Font.Style then
  begin
    Pen.Color := Font.Color;
    if Font.Size > 16 then Pen.Width := 2 else Pen.Width := 1;
    MoveTo(X, Y + fpFnt.Asc + 2);
    LineTo(TextWidth(Text) + X, Y + fpFnt.Asc + 2);
  end;

  if fsStrikeOut in Font.Style then
  begin
    Pen.Color := Font.Color;
    if Font.Size > 16 then Pen.Width := 2 else Pen.Width := 1;
    MoveTo(X, Y + fpFnt.Asc - fpFnt.Desc);
    LineTo(TextWidth(Text) + X, Y + fpFnt.Asc - fpFnt.Desc);
  end;
end;


function TAACanvas.CheckAmpers(var Text: WideString; var Pos: TSize): boolean;
var posit: integer;
begin
  Result := false;
  pos.cx := 0;
  pos.cy := 0;
  posit := System.pos('&&', Text);
  if posit > 0 then
    Delete(Text, posit, 1)
  else
  begin
    posit := System.pos('&', Text);
    if posit < length(Text) then
    begin
      if posit > 0 then
      begin
        Text := Copy(Text, 1, posit - 1) + Copy(Text, posit + 1, length(Text));
        pos.cx := TextWidth(Copy(Text, 1, posit - 1));
        pos.cy := TextWidth(Text[posit]);
        Result := true;
      end;
    end;
  end;
end;

procedure TAACanvas.DoAmperStrike(x, y: integer; Pos: TSize);
begin
  if pos.cy > 0 then
  begin
    Pen.Color := Font.Color;
    if Font.Size > 16 then Pen.Width := 2 else Pen.Width := 1;
    MoveTo(X + Pos.cx, Y + fpFnt.Asc + 2);
    LineTo(X + Pos.cx + Pos.cy, Y + fpFnt.Asc + 2);
  end;
end;


procedure TAACanvas.TextOut(X, Y: Integer; const Text: WideString);
var WorkText: WideString;
  rg: Region;
  Rec: TRect;
begin
  if (not fDontUseAntiAliasing) and (fpFnt <> nil) then
  begin
    WorkText := Text;
    Start;
    Rec := ClipRect;
    rg := QRegion_handle(QRegion_Create(@Rec, QRegionRegionType_Rectangle));
    XftDrawSetClip(fDrawX, rg);
    XFtDrawString16(fDrawX, fFtColor, fpFnt, X, Y + fpFnt.Asc, PWidechar(WorkText), Length(WorkText));
    DoUnderStrike(X, Y, WorkText);
    Stop;
  end else inherited TextOut(X, Y, Text);
end;

procedure TAACanvas.TextRect(Rect: TRect; X, Y: Integer;
  const Text: WideString; TextFlags: Integer);
var rg: Region;
  ix, iy: integer;
  WorkText: WideString;
  AmpPos: TSize;
  outrect: TRect;
begin
  if not fDontUseAntiAliasing and (fpFnt <> nil) then
  begin
    IntersectRect(outrect, Cliprect, Rect);

    WorkText := Text;
    ix := X;
    iy := Y;

    if (TextFlags and Ord(AlignmentFlags_ShowPrefix)) <> 0 then
      CheckAmpers(WorkText, AmpPos);

    if (TextFlags and Ord(AlignmentFlags_AlignRight)) <> 0 then
      ix := Rect.Right - TextWidth(WorkText) - 3 else
      if (TextFlags and Ord(AlignmentFlags_AlignHCenter)) <> 0 then
        ix := ix + (Rect.Right - ix - TextWidth(WorkText) - 3) div 2;

    if (TextFlags and Ord(AlignmentFlags_AlignBottom)) <> 0 then
      iy := Rect.Bottom - TextHeight(WorkText) else
      if (TextFlags and Ord(AlignmentFlags_AlignVCenter)) <> 0 then
        iy := iy + (Rect.Bottom - iy - TextHeight(WorkText)) div 2;
    Start;
    QWMatrix_map(QPainter_worldMatrix(Handle), PRect(@outRect), PRect(@outRect));
    rg := QRegion_handle(QRegion_Create(@outRect, QRegionRegionType_Rectangle));


    XftDrawSetClip(fDrawX, rg);
    XFtDrawString16(fDrawX, fFtColor, fpFnt, iX, iY + fpFnt.Asc, PWidechar(WorkText), Length(WorkText));

    DoUnderStrike(iX, iY, WorkText);
    if (TextFlags and Ord(AlignmentFlags_ShowPrefix)) <> 0 then
      DoAmperStrike(iX, iY, AmpPos);
    Stop;
  end else inherited TextRect(Rect, X, Y, Text, TextFlags);
end;

{ THackControl }

procedure THackControl.AdjustPainter(Painter: QPainterH);
begin
  inherited AdjustPainter(Painter);
end;

function THackControl.GetPaintDevice: QPaintDeviceH;
begin
  Result := inherited GetPaintDevice;
end;

{TAAControlCanvas}

procedure TAAControlCanvas.BeginPainting;
begin
  if not QPainter_isActive(Handle) then
    QPainter_begin(Handle, Control.GetPaintDevice);
  Control.AdjustPainter(Handle);
  DrawNeeded;
end;

procedure TAAControlCanvas.CreateHandle;
begin
  if FControl = nil then
    inherited CreateHandle
  else
    Handle := QPainter_create;
end;

destructor TAAControlCanvas.Destroy;
begin
  FreeHandle;
  inherited Destroy;
end;

procedure TAAControlCanvas.FreeHandle;
var
  H: QPainterH;
begin
  H := FHandle;
  if H <> nil then
  begin
    Handle := nil;
    QPainter_destroy(H);
  end;
end;

procedure TAAControlCanvas.SetClipped(const Value: Boolean);
begin
  if StartCount > 0 then
    raise ECanvasException.CreateRes(@sInvalidSetClipped);
  FClipped := Value;
end;

procedure TAAControlCanvas.SetControl(AControl: THackControl);
begin

  if FControl <> AControl then
  begin
    FreeHandle;
    FControl := AControl;
  end;
  fContainer := fControl;
end;

procedure TAAControlCanvas.StartPaint;
begin
  Start;
end;

procedure TAAControlCanvas.StopPaint;
begin
  Stop;
end;

{ TBitmapCanvas }

procedure InvalidOperation(Str: PResStringRec);
begin
  raise EInvalidGraphicOperation.CreateRes(Str);
end;

procedure TAABitmapCanvas.BeginPainting;
begin
  if not QPainter_isActive(Handle) then
    if not QPainter_begin(Handle, FBitmap.Handle) then
      InvalidOperation(@SInvalidCanvasState);
  DrawNeeded;
end;

constructor TAABitmapCanvas.Create(Bitmap: TAABitmap);
begin
  inherited Create;
  FBitmap := Bitmap;
  fContainer := fBitmap;
end;

procedure TAABitmapCanvas.CreateHandle;
begin
  Handle := QPainter_create;
end;

{ TAABitmap }

procedure TAABitmap.Assign(Source: TPersistent);
begin
  inherited;
  FreeImage;
end;

class function TAABitmap.AssignsTo(Dest: TGraphicClass): Boolean;
begin
  Result := inherited AssignsTo(Dest);
end;

constructor TAABitmap.create;
begin
  inherited;
  FCanvas := TAABitmapCanvas.Create(self);
end;

destructor TAABitmap.destroy;
begin
  FCanvas.free;
  inherited;
end;

function TAABitmap.GetCanvas: TAABitmapCanvas;
begin
  Result := fCanvas;
end;

{ TCustomAAControl }

constructor TCustomAAControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TAAControlCanvas.Create;
  TAAControlCanvas(FCanvas).Control := THackControl(Self);
end;

procedure TCustomAAControl.CreateWidget;
begin
  FHandle := QWidget_create(ParentWidget, nil, WidgetFlags);
  Hooks := QWidget_hook_create(FHandle);
end;

destructor TCustomAAControl.Destroy;
begin
  FCanvas.Free;
  inherited Destroy;
end;

procedure TCustomAAControl.Painting(Sender: QObjectH; EventRegion: QRegionH);
begin
  TAAControlCanvas(FCanvas).StartPaint;
  try
    QPainter_setClipRegion(FCanvas.Handle, EventRegion);
    Paint;
  finally
    TAAControlCanvas(FCanvas).StopPaint;
  end;
end;

procedure TCustomAAControl.Paint;
begin
  { Notification }
end;

procedure TCustomAAControl.DrawMask(Canvas: TCanvas);
var
  R: TRect;
  I: Integer;
begin
  R := ClientRect;
  Canvas.Brush.Color := clMask;
  Canvas.FillRect(R);
  Canvas.Brush.Color := clDontMask;
  if ControlCount > 0 then
  for I := 0 to ControlCount - 1 do
  begin
    if Controls[I] is TWidgetControl then
      Canvas.Brush.Color := clDontMask;
    R := TControl(Controls[I]).BoundsRect;
    Canvas.FillRect(R);
  end;
end;

procedure TCustomAAControl.MaskChanged;
begin
  if Masked then
    UpdateMask
  else
    QWidget_clearMask(Handle);
end;

procedure TCustomAAControl.UpdateMask;
var
  QB: QBitmapH;
  QP: QPainterH;
  Canvas: TCanvas;
begin
  if not Masked or not HandleAllocated then Exit;
  QB := QBitmap_create(Width, Height, True, QPixmapOptimization_DefaultOptim);
  try
    QP := QPainter_create(QB, Handle);
    try
      Canvas := TCanvas.Create;
      try
        Canvas.Start(False);
        Canvas.Handle := QP;
        DrawMask(Canvas);
        Canvas.Stop;
      finally
        Canvas.Free;
      end;
      QWidget_setMask(Handle, QB);
    finally
      QPainter_destroy(QP);
    end;
  finally
    QBitmap_destroy(QB);
  end;
end;

{ TFontCache }

constructor TFontCache.Create;
begin
  fFontInfoList := TList.Create;
end;

destructor TFontCache.Destroy;
var i: integer;
begin
  if fFontInfoList.Count > 0 then
    for i := 0 to fFontInfoList.Count - 1 do
    begin
      TFontCacheItem(fFontInfoList[i]).Font.Free;
      XftFontClose(Application.Display, TFontCacheItem(fFontInfoList[i]).fPfnt);
    end;
  fFontInfoList.Free;
  inherited;
end;

procedure TFontCache.AddFont(ClxFont: QGraphics.TFont; XftFont: PXftFont);
var ci: TFontCacheItem;
begin
  ci := TFontCacheItem.Create;
  ci.Font := QGraphics.TFont.Create;
  ci.Font.Assign(ClxFont);
  ci.fPfnt := XftFont;
  fFontInfoList.Add(ci);
end;


function TFontCache.GetFont(ClxFont: QGraphics.TFont): PXFtFont;
var i: integer;
  ClxFontItem: QGraphics.TFont;
begin
  Result := nil;
  for i := 0 to fFontInfoList.count - 1 do
  begin
    ClxFontItem := TFontCacheItem(fFontInfoList[i]).Font;
    if (ClxFontItem.Name = ClxFont.Name) and
      (ClxFontItem.Style = ClxFont.Style) and
      (ClxFontItem.Weight = ClxFont.Weight) and
      (ClxFontItem.Pitch = ClxFont.Pitch) and      
      (ClxFontItem.Size = ClxFont.Size) then
    begin
      Result := TFontCacheItem(fFontInfoList[i]).fPfnt;
      break;
    end;
  end;
end;


{libXft.so imports}

function XftDrawCreate(display: PDisplay; win: XID; vis: PVisual; colorm: longint): PXftDraw; cdecl; external LibName;

function XftDrawCreateBitmap (display: PDisplay; bitmap: Pixmap): PXftDraw; cdecl; external LibName;

procedure XftDrawChange(xftd: PXftDraw; win: XID); cdecl; external LibName;

procedure XftDrawDestroy(draw: PXftDraw); cdecl; external LibName;

function XftFontOpenName(display: PDisplay; scr: integer; par3: PChar): PXftFont; cdecl; external LibName;

procedure XftFontClose(display: PDisplay; fnt: PXftFont); cdecl; external LibName;

procedure XftDrawString8(draw: PXftDraw; var col: TXftColor; fnt: PXftFont; x, y: integer; txt: PChar; len: integer); cdecl; external LibName;

procedure XftDrawString16(draw: PXftDraw; var col: TXftColor; fnt: PXftFont; x, y: integer; txt: PWideChar; len: integer); cdecl; external LibName;

procedure XftDrawStringUtf8(draw: PXftDraw; var col: TXftColor; fnt: PXftFont; x, y: integer; txt: PChar; len: integer); cdecl; external LibName;

procedure XftTextExtents16(display: PDisplay; fnt: PXftFont; txt: PWideChar; len: integer; var extents: TXGlyphInfo); cdecl; external LibName;

procedure XftTextExtents8(display: PDisplay; fnt: PXftFont; txt: PChar; len: integer; var extents: TXGlyphInfo); cdecl; external LibName;

function XftGlyphExists(display: PDisplay; fnt: PXftFont; ch: integer): longbool; cdecl; external LibName;

procedure XftDrawRect(draw: PXftDraw; var col: TXftColor; x, y: integer; width, height: Cardinal); cdecl; external LibName;

procedure XftDrawSetClip(draw: PXftDraw; rg: Region); cdecl; external LibName;



initialization

  if dlopen('libXft.so', RTLD_LAZY or RTLD_GLOBAL) = nil then
  begin
    WriteLn(ErrOutput, dlerror);
    ExitCode := -1;
    raise EInOutError.Create('libXft.so not found');
    Halt;
  end;


end.
