{******************************************************************************}
{* This software is provided 'as-is', without any express or                  *}
{* implied warranty. In no event will the author be held liable               *}
{* for any damages arising from the use of this software.                     *}
{*                                                                            *}
{* Permission is granted to anyone to use this software for any               *}
{* purpose, including commercial applications, and to alter it                *}
{* and redistribute it freely, subject to the following                       *}
{* restrictions:                                                              *}
{*                                                                            *}
{*   1. The origin of this software must not be misrepresented,               *}
{*      you must not claim that you wrote the original software.              *}
{*      If you use this software in a product, an acknowledgment              *}
{*      in the product documentation would be appreciated but is              *}
{*      not required.                                                         *}
{*                                                                            *}
{*   2. Altered source versions must be plainly marked as such, and           *}
{*      must not be misrepresented as being the original software.            *}
{*                                                                            *}
{*   3. This notice may not be removed or altered from any source             *}
{*      distribution.                                                         *}
{*                                                                            *}
{*                                                                            *}
{* The initial developer is Theo Lustenberger (temp@theo.ch)                  *}
{*                                                                            *}
{******************************************************************************}

// 8.10.2004

unit uAntiAliasLabel;

interface

uses Classes, QTypes, uAntiAliasControl, Qt, QControls, QStdCtrls,
  QGraphics;

type

  TTextLayout = (tlTop, tlCenter, tlBottom);

  TAALabel = class(TCustomAAControl)
  private
    fCaption: TCaption;
    fAlignment: TAlignment;
    fLayout: TTextLayout;
    fAutoSize: Boolean;
    fShowAccelChar: Boolean;
    procedure SetAutoSize(Value: Boolean);
    procedure FontChange(Sender: TObject);
    procedure SetAlignment(Value: TAlignment);
    procedure SetLayout(Value: TTextLayout);
    procedure SetShowAccelChar(Value: Boolean);
  protected
    procedure SetText(const Value: TCaption); override;
    procedure Paint; override;
    procedure InitWidget; override;
    procedure BoundsChanged; override;
    procedure DoAutoSize;
    procedure DrawMask(maCanvas: TCanvas); override;
    procedure PaintToCanvas(aCanvas: TAACanvas);
  public
    constructor Create(AOwner: TComponent); override;
    property Canvas;
  published
    property Caption: TCaption read fCaption write SetText;
    property Alignment: TAlignment read fAlignment write SetAlignment default taLeftJustify;
    property Layout: TTextLayout read fLayout write SetLayout default tlTop;
    property AutoSize: Boolean read FAutoSize write SetAutoSize;
    property ShowAccelChar: Boolean read fShowAccelChar write SetShowAccelChar;
    property Font stored true;
    property ParentFont;
    property Bitmap;
    property Color;
    property ParentColor;
    property Align;
    property Anchors;
    property Constraints;
    property DragMode;
    property Enabled;
    property ParentShowHint;
    property Popupmenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property Masked;
  end;

procedure Register;


implementation


{ TAALabel }

procedure TAALabel.PaintToCanvas(aCanvas: TAACanvas);
var AlFlags: integer;
begin
  inherited;

  AlFlags := 0;

  if fShowAccelChar then
    AlFlags := AlFlags or Ord(AlignmentFlags_ShowPrefix);

  if not fAutosize then
  begin
    case fAlignment of
      taLeftJustify: ;
      taRightJustify: AlFlags := AlFlags or Ord(AlignmentFlags_AlignRight);
      taCenter: AlFlags := AlFlags or Ord(AlignmentFlags_AlignHCenter);
    end;

    case fLayout of
      tlTop: ;
      tlCenter: AlFlags := AlFlags or Ord(AlignmentFlags_AlignVCenter);
      tlBottom: AlFlags := AlFlags or Ord(AlignmentFlags_AlignBottom);
    end;

    aCanvas.TextRect(Rect(0, 0, width, height), 0, 0, fCaption, AlFlags);
  end else aCanvas.TextRect(Rect(0, 0, width, height), 0, 0, fCaption, AlFlags);
end;

procedure TAALabel.Paint;

begin
  inherited;
  PaintToCanvas(Canvas);
end;

procedure TAALabel.DrawMask(maCanvas: TCanvas);

type
  PRGBQuad = ^TRGBQuad;
  TRGBQuad = packed record
    rgbBlue: Byte;
    rgbGreen: Byte;
    rgbRed: Byte;
    rgbReserved: Byte;
  end;

  PRGBQuadArray = ^TRGBQuadArray;
  TRGBQuadArray = array[0..MaxInt div SizeOf(TRGBQuad) - 1] of TRGBQuad;


var bmp: TAABitmap;
  BX, BY: Integer;
  TransparentColor: TRGBQuad;
  Line: PRGBQuadArray;
  Pts: TPointArray;
  i: integer;
begin
  inherited;
  bmp := TAABitmap.Create;
  bmp.Width := self.Width;
  bmp.Height := self.Height;
  SetLength(Pts, bmp.Width * bmp.Height);

  TransparentColor.rgbRed := $FF;
  TransparentColor.rgbGreen := $FF;
  TransparentColor.rgbBlue := $FF;

  bmp.Canvas.Start;
  bmp.Canvas.Brush.Color := cldontmask;
  bmp.Canvas.FillRect(Rect(0, 0, width, height));
  bmp.Canvas.Font.Assign(Font);
  bmp.Canvas.Font.Color := clmask;
  PaintToCanvas(bmp.Canvas);
  bmp.Canvas.stop;

  i := 0;
  for BY := 0 to Height - 1 do
  begin
    Line := pRGBQuadArray(bmp.Scanline[BY]);
    for BX := 0 to Width - 1 do
      if (Cardinal(Line[BX]) and $000000FF) <>
        (Cardinal(TransparentColor) and $000000FF) then
      begin
        Pts[i] := Point(BX, BY);
        inc(i);
      end;
  end;
  SetLength(pts, i);

  if i > 0 then
  try
    maCanvas.Start(false);
    maCanvas.Pen.Color := cldontmask;
    maCanvas.DrawPoints(Pts);
  finally
    maCanvas.Stop;
  end;

  bmp.free;
end;


procedure TAALabel.SetText(const Value: TCaption);
begin
  if Value <> fCaption then
  begin
    fCaption := Value;
    Invalidate;
    DoAutoSize;
  end;
end;

procedure TAALabel.SetAutoSize(Value: Boolean);
begin
  if Value <> fAutoSize then
  begin
    FAutoSize := Value;
    Invalidate;
    DoAutoSize;
  end;
end;

constructor TAALabel.Create(AOwner: TComponent);
begin
  inherited;
  if [csDesigning, csLoading] * Aowner.ComponentState = [csDesigning] then
  begin
    fAutoSize := True;
    Caption := Name;
  end;
  Font.OnChange := FontChange;
end;

procedure TAALabel.InitWidget;
begin
  inherited;
  QWidget_setFocusPolicy(Handle, QWidgetFocusPolicy_NoFocus);
end;

procedure TAALabel.DoAutoSize;
begin
  if FAutoSize then
  begin
    Align := alNone;
    Width := Canvas.TextWidth(fCaption);
    Height := Canvas.TextHeight(fCaption);
  end;
end;

procedure TAALabel.FontChange(Sender: TObject);
begin
  if not (csLoading in ComponentState) then ParentFont := false;
  Canvas.Font.Assign(Font);
  DoAutoSize;
  Invalidate;
end;

procedure TAALabel.SetAlignment(Value: TAlignment);
begin
  fAlignment := value;
  Invalidate;
end;

procedure TAALabel.SetLayout(Value: TTextLayout);
begin
  fLayout := Value;
  Invalidate;
end;


procedure TAALabel.SetShowAccelChar(Value: Boolean);
begin
  fShowAccelChar := Value;
  Invalidate;
end;

procedure TAALabel.BoundsChanged;
begin
  inherited BoundsChanged;
  DoAutoSize;
end;

procedure Register;
begin
  RegisterComponents('Theos', [TAALabel]);
end;

end.
